@@ -4,23 +4,29 @@ __all__ = [
     'create_session',
 ]
 
-__version__ = '6.7.0'
+# This should always be 0.0.0 in master. Only update this after tagging
+# before release.
+__version__ = '9.5.0'
 
 import contextlib
 import json
 import logging
-import os
 import random
-import six
 import time
 
 import requests
+import six
 
-from . import stone_serializers
+from . import files, stone_serializers
 from .auth import (
     AuthError_validator,
     RateLimitError_validator,
 )
+from .common import (
+    PathRoot,
+    PathRoot_validator,
+    PathRootError_validator
+)
 from .base import DropboxBase
 from .base_team import DropboxTeamBase
 from .exceptions import (
@@ -28,11 +34,22 @@ from .exceptions import (
     AuthError,
     BadInputError,
     HttpError,
+    PathRootError,
     InternalServerError,
     RateLimitError,
 )
-from .session import pinned_session
+from .session import (
+    API_HOST,
+    API_CONTENT_HOST,
+    API_NOTIFICATION_HOST,
+    HOST_API,
+    HOST_CONTENT,
+    HOST_NOTIFY,
+    pinned_session,
+)
 
+PATH_ROOT_HEADER = 'Dropbox-API-Path-Root'
+HTTP_STATUS_INVALID_PATH_ROOT = 422
 
 class RouteResult(object):
     """The successful result of a call to a route."""
@@ -53,7 +70,6 @@ class RouteResult(object):
         self.obj_result = obj_result
         self.http_resp = http_resp
 
-
 class RouteErrorResult(object):
     """The error result of a call to a route."""
 
@@ -67,14 +83,12 @@ class RouteErrorResult(object):
         self.request_id = request_id
         self.obj_result = obj_result
 
-
 def create_session(max_connections=8, proxies=None):
     """
     Creates a session object that can be used by multiple :class:`Dropbox` and
     :class:`DropboxTeam` instances. This lets you share a connection pool
     amongst them, as well as proxy parameters.
 
-
     :param int max_connections: Maximum connection pool size.
     :param dict proxies: See the `requests module
             <http://docs.python-requests.org/en/latest/user/advanced/#proxies>`_
@@ -89,7 +103,6 @@ def create_session(max_connections=8, proxies=None):
         session.proxies = proxies
     return session
 
-
 class _DropboxTransport(object):
     """
     Responsible for implementing the wire protocol for making requests to the
@@ -98,17 +111,6 @@ class _DropboxTransport(object):
 
     _API_VERSION = '2'
 
-    _DEFAULT_DOMAIN = '.dropboxapi.com'
-
-    # Host for RPC-style routes.
-    _HOST_API = 'api'
-
-    # Host for upload and download-style routes.
-    _HOST_CONTENT = 'content'
-
-    # Host for longpoll routes.
-    _HOST_NOTIFY = 'notify'
-
     # Download style means that the route argument goes in a Dropbox-API-Arg
     # header, and the result comes back in a Dropbox-API-Result header. The
     # HTTP response body contains a binary payload.
@@ -123,13 +125,17 @@ class _DropboxTransport(object):
     # the HTTP body.
     _ROUTE_STYLE_RPC = 'rpc'
 
+    # This is the default longest time we'll block on receiving data from the server
+    _DEFAULT_TIMEOUT = 100
+
     def __init__(self,
                  oauth2_access_token,
                  max_retries_on_error=4,
                  max_retries_on_rate_limit=None,
                  user_agent=None,
                  session=None,
-                 headers=None):
+                 headers=None,
+                 timeout=_DEFAULT_TIMEOUT):
         """
         :param str oauth2_access_token: OAuth2 access token for making client
             requests.
@@ -147,6 +153,11 @@ class _DropboxTransport(object):
             :func:`create_session`.
         :type session: :class:`requests.sessions.Session`
         :param dict headers: Additional headers to add to requests.
+        :param Optional[float] timeout: Maximum duration in seconds that
+            client will wait for any single packet from the
+            server. After the timeout the client will give up on
+            connection. If `None`, client will wait forever. Defaults
+            to 30 seconds.
         """
         assert len(oauth2_access_token) > 0, \
             'OAuth2 access token cannot be empty.'
@@ -174,22 +185,47 @@ class _DropboxTransport(object):
 
         self._logger = logging.getLogger('dropbox')
 
-        self._domain = os.environ.get('DROPBOX_DOMAIN', Dropbox._DEFAULT_DOMAIN)
-        self._api_hostname = os.environ.get(
-            'DROPBOX_API_HOST', 'api' + self._domain)
-        self._api_content_hostname = os.environ.get(
-            'DROPBOX_API_CONTENT_HOST', 'content' + self._domain)
-        self._api_notify_hostname = os.environ.get(
-            'DROPBOX_API_NOTIFY_HOST', 'notify' + self._domain)
-        self._host_map = {self._HOST_API: self._api_hostname,
-                          self._HOST_CONTENT: self._api_content_hostname,
-                          self._HOST_NOTIFY: self._api_notify_hostname}
+        self._host_map = {HOST_API: API_HOST,
+                          HOST_CONTENT: API_CONTENT_HOST,
+                          HOST_NOTIFY: API_NOTIFICATION_HOST}
+
+        self._timeout = timeout
+
+    def clone(
+            self,
+            oauth2_access_token=None,
+            max_retries_on_error=None,
+            max_retries_on_rate_limit=None,
+            user_agent=None,
+            session=None,
+            headers=None,
+            timeout=None):
+        """
+        Creates a new copy of the Dropbox client with the same defaults unless modified by
+        arguments to clone()
+
+        See constructor for original parameter descriptions.
+
+        :return: New instance of Dropbox clent
+        :rtype: Dropbox
+        """
+
+        return self.__class__(
+            oauth2_access_token or self._oauth2_access_token,
+            max_retries_on_error or self._max_retries_on_error,
+            max_retries_on_rate_limit or self._max_retries_on_rate_limit,
+            user_agent or self._user_agent,
+            session or self._session,
+            headers or self._headers,
+            timeout or self._timeout
+        )
 
     def request(self,
                 route,
                 namespace,
                 request_arg,
-                request_binary):
+                request_binary,
+                timeout=None):
         """
         Makes a request to the Dropbox API and in the process validates that
         the route argument and result are the expected data types. The
@@ -204,18 +240,38 @@ class _DropboxTransport(object):
             validator specified by route.arg_type.
         :param request_binary: String or file pointer representing the binary
             payload. Use None if there is no binary payload.
+        :param Optional[float] timeout: Maximum duration in seconds
+            that client will wait for any single packet from the
+            server. After the timeout the client will give up on
+            connection. If `None`, will use default timeout set on
+            Dropbox object.  Defaults to `None`.
         :return: The route's result.
         """
         host = route.attrs['host'] or 'api'
         route_name = namespace + '/' + route.name
+        if route.version > 1:
+            route_name += '_v{}'.format(route.version)
         route_style = route.attrs['style'] or 'rpc'
         serialized_arg = stone_serializers.json_encode(route.arg_type,
                                                        request_arg)
+
+        if (timeout is None and
+                route == files.list_folder_longpoll):
+            # The client normally sends a timeout value to the
+            # longpoll route. The server will respond after
+            # <timeout> + random(0, 90) seconds. We increase the
+            # socket timeout to the longpoll timeout value plus 90
+            # seconds so that we don't cut the server response short
+            # due to a shorter socket timeout.
+            # NB: This is done here because base.py is auto-generated
+            timeout = request_arg.timeout + 90
+
         res = self.request_json_string_with_retry(host,
                                                   route_name,
                                                   route_style,
                                                   serialized_arg,
-                                                  request_binary)
+                                                  request_binary,
+                                                  timeout=timeout)
         decoded_obj_result = json.loads(res.obj_result)
         if isinstance(res, RouteResult):
             returned_data_type = route.result_type
@@ -225,12 +281,11 @@ class _DropboxTransport(object):
             obj = decoded_obj_result['error']
             user_message = decoded_obj_result.get('user_message')
             user_message_text = user_message and user_message.get('text')
-            user_message_locale =  user_message and user_message.get('locale')
+            user_message_locale = user_message and user_message.get('locale')
         else:
             raise AssertionError('Expected RouteResult or RouteErrorResult, '
                                  'but res is %s' % type(res))
 
-
         deserialized_result = stone_serializers.json_compat_obj_decode(
             returned_data_type, obj, strict=False)
 
@@ -249,7 +304,8 @@ class _DropboxTransport(object):
                             route_name,
                             route_style,
                             request_arg,
-                            request_binary):
+                            request_binary,
+                            timeout=None):
         """
         Makes a request to the Dropbox API, taking a JSON-serializable Python
         object as an argument, and returning one as a response.
@@ -259,8 +315,13 @@ class _DropboxTransport(object):
         :param route_style: The style of the route.
         :param str request_arg: A JSON-serializable Python object representing
             the argument for the route.
-        :param request_binary: String or file pointer representing the binary
+        :param Optional[bytes] request_binary: Bytes representing the binary
             payload. Use None if there is no binary payload.
+        :param Optional[float] timeout: Maximum duration in seconds
+            that client will wait for any single packet from the
+            server. After the timeout the client will give up on
+            connection. If `None`, will use default timeout set on
+            Dropbox object.  Defaults to `None`.
         :return: The route's result as a JSON-serializable Python object.
         """
         serialized_arg = json.dumps(request_arg)
@@ -268,7 +329,8 @@ class _DropboxTransport(object):
                                                   route_name,
                                                   route_style,
                                                   serialized_arg,
-                                                  request_binary)
+                                                  request_binary,
+                                                  timeout=timeout)
         # This can throw a ValueError if the result is not deserializable,
         # but that would be completely unexpected.
         deserialized_result = json.loads(res.obj_result)
@@ -282,7 +344,8 @@ class _DropboxTransport(object):
                                        route_name,
                                        route_style,
                                        request_json_arg,
-                                       request_binary):
+                                       request_binary,
+                                       timeout=None):
         """
         See :meth:`request_json_object` for description of parameters.
 
@@ -298,7 +361,8 @@ class _DropboxTransport(object):
                                                 route_name,
                                                 route_style,
                                                 request_json_arg,
-                                                request_binary)
+                                                request_binary,
+                                                timeout=timeout)
             except InternalServerError as e:
                 attempt += 1
                 if attempt <= self._max_retries_on_error:
@@ -327,7 +391,8 @@ class _DropboxTransport(object):
                             func_name,
                             route_style,
                             request_json_arg,
-                            request_binary):
+                            request_binary,
+                            timeout=None):
         """
         See :meth:`request_json_string_with_retry` for description of
         parameters.
@@ -335,12 +400,20 @@ class _DropboxTransport(object):
         if host not in self._host_map:
             raise ValueError('Unknown value for host: %r' % host)
 
+        if not isinstance(request_binary, (six.binary_type, type(None))):
+            # Disallow streams and file-like objects even though the underlying
+            # requests library supports them. This is to prevent incorrect
+            # behavior when a non-rewindable stream is read from, but the
+            # request fails and needs to be re-tried at a later time.
+            raise TypeError('expected request_binary as binary type, got %s' %
+                            type(request_binary))
+
         # Fully qualified hostname
         fq_hostname = self._host_map[host]
         url = self._get_route_url(fq_hostname, func_name)
 
         headers = {'User-Agent': self._user_agent}
-        if host != self._HOST_NOTIFY:
+        if host != HOST_NOTIFY:
             headers['Authorization'] = 'Bearer %s' % self._oauth2_access_token
             if self._headers:
                 headers.update(self._headers)
@@ -365,11 +438,15 @@ class _DropboxTransport(object):
         else:
             raise ValueError('Unknown operation style: %r' % route_style)
 
+        if timeout is None:
+            timeout = self._timeout
+
         r = self._session.post(url,
                                headers=headers,
                                data=body,
                                stream=stream,
                                verify=True,
+                               timeout=timeout,
                                )
 
         request_id = r.headers.get('x-dropbox-request-id')
@@ -384,6 +461,10 @@ class _DropboxTransport(object):
             err = stone_serializers.json_compat_obj_decode(
                 AuthError_validator, r.json()['error'])
             raise AuthError(request_id, err)
+        elif r.status_code == HTTP_STATUS_INVALID_PATH_ROOT:
+            err = stone_serializers.json_compat_obj_decode(
+                PathRootError_validator, r.json()['error'])
+            raise PathRootError(request_id, err)
         elif r.status_code == 429:
             err = None
             if r.headers.get('content-type') == 'application/json':
@@ -442,6 +523,27 @@ class _DropboxTransport(object):
                 for c in http_resp.iter_content(chunksize):
                     f.write(c)
 
+    def with_path_root(self, path_root):
+        """
+        Creates a clone of the Dropbox instance with the Dropbox-API-Path-Root header
+        as the appropriate serialized instance of PathRoot.
+
+        For more information, see
+        https://www.dropbox.com/developers/reference/namespace-guide#pathrootmodes
+
+        :param PathRoot path_root: instance of PathRoot to serialize into the headers field
+        :return: A :class: `Dropbox`
+        :rtype: Dropbox
+        """
+
+        if not isinstance(path_root, PathRoot):
+            raise ValueError("path_root must be an instance of PathRoot")
+
+        return self.clone(
+            headers={
+                PATH_ROOT_HEADER: stone_serializers.json_encode(PathRoot_validator, path_root)
+            }
+        )
 
 class Dropbox(_DropboxTransport, DropboxBase):
     """
@@ -451,29 +553,56 @@ class Dropbox(_DropboxTransport, DropboxBase):
     """
     pass
 
-
 class DropboxTeam(_DropboxTransport, DropboxTeamBase):
     """
     Use this class to make requests to the Dropbox API using a team's access
     token. Methods of this class are meant to act on the team, but there is
     also an :meth:`as_user` method for assuming a team member's identity.
     """
+    def as_admin(self, team_member_id):
+        """
+        Allows a team credential to assume the identity of an administrator on the team
+        and perform operations on any team-owned content.
+
+        :param str team_member_id: team member id of administrator to perform actions with
+        :return: A :class:`Dropbox` object that can be used to query on behalf
+            of this admin of the team.
+        :rtype: Dropbox
+        """
+        return self._get_dropbox_client_with_select_header('Dropbox-API-Select-Admin',
+                                                           team_member_id)
 
     def as_user(self, team_member_id):
         """
         Allows a team credential to assume the identity of a member of the
         team.
 
+        :param str team_member_id: team member id of team member to perform actions with
         :return: A :class:`Dropbox` object that can be used to query on behalf
             of this member of the team.
         :rtype: Dropbox
         """
+        return self._get_dropbox_client_with_select_header('Dropbox-API-Select-User',
+                                                           team_member_id)
+
+    def _get_dropbox_client_with_select_header(self, select_header_name, team_member_id):
+        """
+        Get Dropbox client with modified headers
+
+        :param str select_header_name: Header name used to select users
+        :param str team_member_id: team member id of team member to perform actions with
+        :return: A :class:`Dropbox` object that can be used to query on behalf
+            of a member or admin of the team
+        :rtype: Dropbox
+        """
+
         new_headers = self._headers.copy() if self._headers else {}
-        new_headers['Dropbox-API-Select-User'] = team_member_id
+        new_headers[select_header_name] = team_member_id
         return Dropbox(
             self._oauth2_access_token,
             max_retries_on_error=self._max_retries_on_error,
             max_retries_on_rate_limit=self._max_retries_on_rate_limit,
+            timeout=self._timeout,
             user_agent=self._raw_user_agent,
             session=self._session,
             headers=new_headers,
